#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <app/Roster.h>
#include <interface/Alert.h>
#include <interface/Screen.h>
#include <interface/Rect.h>
#include <interface/View.h>
#include <storage/Directory.h>
#include <storage/Entry.h>
#include <support/List.h>
#include <support/Beep.h>

#include "CommandActuators.h"
#include "ParseCommandLine.h"
#include "KeyInfos.h"

#define IS_KEY_DOWN(msg) ((msg->what == B_KEY_DOWN)||(msg->what == B_UNMAPPED_KEY_DOWN))

// Factory function
CommandActuator * CreateCommandActuator(const char * command)
{   
    CommandActuator * act = NULL; 
    int32 argc;
    char ** argv = ParseArgvFromString(command, argc);       
    if (command[0] == '*')
    {
       if (argc > 0)
       {
          char * c = argv[0]+1;
               if (strcmp(c, "InsertString") == 0) act = new KeyStrokeSequenceCommandActuator(argc, argv);
          else if (strcmp(c, "MoveMouse")    == 0) act = new MoveMouseByCommandActuator(argc, argv);
          else if (strcmp(c, "MoveMouseTo")  == 0) act = new MoveMouseToCommandActuator(argc, argv);
          else if (strcmp(c, "MouseButton")  == 0) act = new MouseButtonCommandActuator(argc, argv);
          else if (strcmp(c, "LaunchHandler")== 0) act = new MIMEHandlerCommandActuator(argc, argv);  // new for v1.21 --jaf
          else if (strcmp(c, "Multi")        == 0) act = new MultiCommandActuator(argc, argv);        // new for v1.24 --jaf
          else if (strcmp(c, "MouseDown")    == 0) act = new MouseDownCommandActuator(argc, argv);
          else if (strcmp(c, "MouseUp")      == 0) act = new MouseUpCommandActuator(argc, argv);
          else if (strcmp(c, "SendMessage")  == 0) act = new SendMessageCommandActuator(argc, argv);  // new for v1.25 --jaf
          else                                     act = new BeepCommandActuator(argc, argv);
       } 
    }
    else act = new LaunchCommandActuator(argc, argv);
    
    FreeArgv(argv);
    return act;
}


///////////////////////////////////////////////////////////////////////////////
//
//  CommandActuator
//
///////////////////////////////////////////////////////////////////////////////

CommandActuator :: CommandActuator(int32 argc, char ** argv)
{
   // empty
}

CommandActuator :: CommandActuator(BMessage * from)
   : BArchivable(from)
{
   // empty
}

status_t CommandActuator :: Archive(BMessage * into, bool deep) const
{
   status_t ret = BArchivable::Archive(into, deep);
   return ret;
}


///////////////////////////////////////////////////////////////////////////////
//
//  LaunchCommandActuator
//
///////////////////////////////////////////////////////////////////////////////

LaunchCommandActuator :: LaunchCommandActuator(int32 argc, char ** argv)
   : CommandActuator(argc, argv), _argv(CloneArgv(argv)), _argc(argc)
{
   // empty
}

LaunchCommandActuator :: LaunchCommandActuator(BMessage * from)
   : CommandActuator(from)
{
   BList argList;
   const char * temp;
   int idx = 0;
   while(from->FindString("largv", idx++, &temp) == B_NO_ERROR) 
   {
      if (temp) 
      {
         char * copy = new char[strlen(temp)+1];         
         strcpy(copy, temp);         
         argList.AddItem(copy);
      }      
   }

   _argc = argList.CountItems();
   _argv = new char *[_argc+1];
   for (int i=0; i<_argc; i++) _argv[i] = (char *) argList.ItemAt(i);
   _argv[_argc] = NULL;  // terminate the array
}

LaunchCommandActuator :: ~LaunchCommandActuator()
{
   FreeArgv(_argv);
}

filter_result LaunchCommandActuator :: KeyEvent(const BMessage * keyMsg, BList * outlist, void ** setAsyncData, BMessage * lastMouseMove)
{
   if (IS_KEY_DOWN(keyMsg)) *setAsyncData = (void *) true;  // cause KeyEventAsync() to be called asynchronously
   return B_SKIP_MESSAGE;
}

status_t LaunchCommandActuator :: Archive(BMessage * into, bool deep) const
{
   status_t ret = CommandActuator :: Archive(into, deep);
   for (int i=0; i<_argc; i++) into->AddString("largv", _argv[i]);
   return ret;
}


BArchivable * LaunchCommandActuator ::Instantiate(BMessage * from)
{
   return (validate_instantiation(from, "LaunchCommandActuator")) ? new LaunchCommandActuator(from) : NULL;
}


void LaunchCommandActuator :: KeyEventAsync(const BMessage * keyMsg, void * asyncData)
{
   if (be_roster)
   {
      if (_argc < 1)
      {
         (new BAlert("SpicyKeys Launcher Error", "You didn't specify a command for this hotkey.", "Okay"))->Go(NULL);
      }
      else if (LaunchCommand(_argv, _argc) != B_NO_ERROR)
      {
         const char * str1 = "Can't launch ";
         const char * str2 = ", no such file exists.  Please check your SpicyKeys settings.";
         char * temp = new char[strlen(str1) + strlen(_argv[0]) + strlen(str2) + 1];
         sprintf(temp, "%s%s%s", str1, _argv[0], str2);
         (new BAlert("SpicyKeys Launcher Error", temp, "Okay"))->Go(NULL);      
         delete [] temp;      
      }
   }
}


///////////////////////////////////////////////////////////////////////////////
//
//  MouseCommandActuator
//
///////////////////////////////////////////////////////////////////////////////

MouseCommandActuator :: MouseCommandActuator(int32 argc, char ** argv)
   : CommandActuator(argc, argv), _whichButtons(B_PRIMARY_MOUSE_BUTTON)
{
   if (argc > 1)
   {
      _whichButtons = 0;
      for (int i=1; i<argc; i++)
      {
         int buttonNumber = atoi(argv[i]);
         switch(buttonNumber)
         {
            case 1: _whichButtons |= B_PRIMARY_MOUSE_BUTTON;   break;
            case 2: _whichButtons |= B_SECONDARY_MOUSE_BUTTON; break;
            case 3: _whichButtons |= B_TERTIARY_MOUSE_BUTTON;  break;
         }
      }
   }
}

MouseCommandActuator :: MouseCommandActuator(BMessage * from)
   : CommandActuator(from), _whichButtons(B_PRIMARY_MOUSE_BUTTON)
{
   from->FindInt32("buttons", &_whichButtons);
}

MouseCommandActuator :: ~MouseCommandActuator()
{
   // empty
}

status_t MouseCommandActuator :: Archive(BMessage * into, bool deep) const
{
   status_t ret = CommandActuator :: Archive(into, deep);
   into->AddInt32("buttons", _whichButtons);   
   return ret;
}

int32 MouseCommandActuator :: GetWhichButtons() const
{
   return _whichButtons;
}

void MouseCommandActuator :: GenerateMouseButtonEvent(bool mouseDown, const BMessage * keyMsg, BList * outlist, BMessage * lastMouseMove)
{   
   BMessage *fakeMouse = new BMessage(*lastMouseMove);
   fakeMouse->what = mouseDown ? B_MOUSE_DOWN : B_MOUSE_UP;
   
   // Update the buttons to reflect which mouse buttons we are faking
   fakeMouse->RemoveName("buttons");
   if (mouseDown) fakeMouse->AddInt32("buttons", _whichButtons);   

   // Trey sez you gotta keep then "when"'s increasing if you want click&drag to work!
   int64 when;
   const BMessage * lastMessage = (outlist->CountItems() > 0) ? (const BMessage *)outlist->ItemAt(outlist->CountItems()-1) : keyMsg;
   if (lastMessage->FindInt64("when", &when) == B_NO_ERROR)
   {
      when++;
      fakeMouse->RemoveName("when");
      fakeMouse->AddInt64("when", when);
   }
   
   outlist->AddItem(fakeMouse);
}



///////////////////////////////////////////////////////////////////////////////
//
//  MouseDownCommandActuator
//
///////////////////////////////////////////////////////////////////////////////

MouseDownCommandActuator :: MouseDownCommandActuator(int32 argc, char ** argv)
   : MouseCommandActuator(argc, argv)
{
   // empty
}

MouseDownCommandActuator :: MouseDownCommandActuator(BMessage * from)
   : MouseCommandActuator(from)
{
   // empty
}

MouseDownCommandActuator :: ~MouseDownCommandActuator()
{
   // empty
}

filter_result MouseDownCommandActuator :: KeyEvent(const BMessage * keyMsg, BList * outlist, void ** setAsyncData, BMessage * lastMouseMove)
{
   if (IS_KEY_DOWN(keyMsg)) GenerateMouseButtonEvent(true, keyMsg, outlist, lastMouseMove);
   return B_DISPATCH_MESSAGE;
}

status_t MouseDownCommandActuator :: Archive(BMessage * into, bool deep) const
{
   status_t ret = MouseCommandActuator :: Archive(into, deep);
   return ret;
}

BArchivable * MouseDownCommandActuator ::Instantiate(BMessage * from)
{
   return (validate_instantiation(from, "MouseDownCommandActuator")) ? new MouseDownCommandActuator(from) : NULL;
}

///////////////////////////////////////////////////////////////////////////////
//
//  MouseUpCommandActuator
//
///////////////////////////////////////////////////////////////////////////////

MouseUpCommandActuator :: MouseUpCommandActuator(int32 argc, char ** argv)
   : MouseCommandActuator(argc, argv)
{
   // empty
}

MouseUpCommandActuator :: MouseUpCommandActuator(BMessage * from)
   : MouseCommandActuator(from)
{
   // empty
}

MouseUpCommandActuator :: ~MouseUpCommandActuator()
{
   // empty
}

filter_result MouseUpCommandActuator :: KeyEvent(const BMessage * keyMsg, BList * outlist, void ** setAsyncData, BMessage * lastMouseMove)
{
   if (IS_KEY_DOWN(keyMsg)) GenerateMouseButtonEvent(false, keyMsg, outlist, lastMouseMove);
   return B_DISPATCH_MESSAGE;
}

status_t MouseUpCommandActuator :: Archive(BMessage * into, bool deep) const
{
   status_t ret = MouseCommandActuator :: Archive(into, deep);
   return ret;
}

BArchivable * MouseUpCommandActuator ::Instantiate(BMessage * from)
{
   return (validate_instantiation(from, "MouseUpCommandActuator")) ? new MouseUpCommandActuator(from) : NULL;
}



///////////////////////////////////////////////////////////////////////////////
//
//  MouseButtonCommandActuator
//
///////////////////////////////////////////////////////////////////////////////

MouseButtonCommandActuator :: MouseButtonCommandActuator(int32 argc, char ** argv)
   : MouseCommandActuator(argc, argv), _keyDown(false)
{
   // empty
}

MouseButtonCommandActuator :: MouseButtonCommandActuator(BMessage * from)
   : MouseCommandActuator(from), _keyDown(false)
{
   // empty
}

MouseButtonCommandActuator :: ~MouseButtonCommandActuator()
{
   // empty
}

filter_result MouseButtonCommandActuator :: KeyEvent(const BMessage * keyMsg, BList * outlist, void ** setAsyncData, BMessage * lastMouseMove)
{
   if (IS_KEY_DOWN(keyMsg) != _keyDown)
   {
      GenerateMouseButtonEvent(IS_KEY_DOWN(keyMsg), keyMsg, outlist, lastMouseMove);   
      _keyDown = IS_KEY_DOWN(keyMsg);
      return B_DISPATCH_MESSAGE;
   }
   else return B_SKIP_MESSAGE;  // This will handle key-repeats, which we don't want turned into lots of B_MOUSE_DOWN messages.
}

status_t MouseButtonCommandActuator :: Archive(BMessage * into, bool deep) const
{
   status_t ret = MouseCommandActuator :: Archive(into, deep);
   return ret;
}


BArchivable * MouseButtonCommandActuator ::Instantiate(BMessage * from)
{
   return (validate_instantiation(from, "MouseButtonCommandActuator")) ? new MouseButtonCommandActuator(from) : NULL;
}


///////////////////////////////////////////////////////////////////////////////
//
//  KeyStrokeSequenceCommandActuator
//
///////////////////////////////////////////////////////////////////////////////

KeyStrokeSequenceCommandActuator :: KeyStrokeSequenceCommandActuator(int32 argc, char ** argv)
   : CommandActuator(argc, argv)
{
   for (int s=1; s<argc; s++) 
   {
      _sequence.Append(argv[s]);
      if (s < argc-1) _sequence.Append(" ");
   }
   GenerateKeyCodes();
}

KeyStrokeSequenceCommandActuator :: KeyStrokeSequenceCommandActuator(BMessage * from)
   : CommandActuator(from)
{
   const char * temp;
   if (from->FindString("sequence", 0, &temp) == B_NO_ERROR) _sequence = temp;   
   GenerateKeyCodes();
}

KeyStrokeSequenceCommandActuator :: ~KeyStrokeSequenceCommandActuator()
{
   delete [] _keyCodes;
   delete [] _modCodes;
   delete [] _states;
}

void KeyStrokeSequenceCommandActuator :: GenerateKeyCodes()
{
   int slen = _sequence.Length();
   _keyCodes = new int32[slen];
   _modCodes = new int32[slen];
   _states   = new uint8[slen*16];
   
   key_map * map;
   char * keys;
   get_key_map(&map, &keys);
   for (int i=0; i<slen; i++) 
   {
      // Gotta do reverse-lookups to find out the raw keycodes for a given character.
      // Expensive--there oughtta be a better way to do this.
      char next = _sequence.ByteAt(i);
      _modCodes[i] = 0;
      int32 & mod = _modCodes[i];
      uint8 * states = &_states[i*16];
      int32        key = LookupKeyCode(map, keys, map->normal_map,            next, states, mod, 0);
      if (key < 0) key = LookupKeyCode(map, keys, map->shift_map,             next, states, mod, B_LEFT_SHIFT_KEY | B_SHIFT_KEY);
      if (key < 0) key = LookupKeyCode(map, keys, map->caps_map,              next, states, mod, B_CAPS_LOCK);
      if (key < 0) key = LookupKeyCode(map, keys, map->caps_shift_map,        next, states, mod, B_LEFT_SHIFT_KEY | B_SHIFT_KEY | B_CAPS_LOCK);
      if (key < 0) key = LookupKeyCode(map, keys, map->option_map,            next, states, mod, B_LEFT_OPTION_KEY | B_OPTION_KEY);
      if (key < 0) key = LookupKeyCode(map, keys, map->option_shift_map,      next, states, mod, B_LEFT_OPTION_KEY | B_OPTION_KEY | B_LEFT_SHIFT_KEY | B_SHIFT_KEY);
      if (key < 0) key = LookupKeyCode(map, keys, map->option_caps_map,       next, states, mod, B_LEFT_OPTION_KEY | B_OPTION_KEY | B_CAPS_LOCK);
      if (key < 0) key = LookupKeyCode(map, keys, map->option_caps_shift_map, next, states, mod, B_LEFT_OPTION_KEY | B_OPTION_KEY | B_CAPS_LOCK | B_LEFT_SHIFT_KEY | B_SHIFT_KEY);
      if (key < 0) key = LookupKeyCode(map, keys, map->control_map,           next, states, mod, B_CONTROL_KEY);      
      _keyCodes[i] = key;
   }
}

int32 KeyStrokeSequenceCommandActuator :: LookupKeyCode(key_map * map, char * keys, int32 offsets[128], char c, uint8 * setStates, int32 & setMod, int32 setTo) const
{
   for (int i=0; i<128; i++) 
   {
      if (keys[offsets[i]+1] == c) 
      {
         SetStateBit(setStates, i);
         if (setTo & B_SHIFT_KEY)   SetStateBit(setStates, map->left_shift_key);
         if (setTo & B_OPTION_KEY)  SetStateBit(setStates, map->left_option_key);
         if (setTo & B_CONTROL_KEY) SetStateBit(setStates, map->left_control_key);
         if (setTo & B_CAPS_LOCK)   SetStateBit(setStates, map->caps_key);
         setMod = setTo;
         return i;
      }
   }
   return -1;
}

void KeyStrokeSequenceCommandActuator :: SetStateBit(uint8 * setStates, uint32 key) const
{
   setStates[key/8] |= (0x80 >> (key%8));
}

status_t KeyStrokeSequenceCommandActuator :: Archive(BMessage * into, bool deep) const
{
   status_t ret = CommandActuator :: Archive(into, deep);
   into->AddString("sequence", _sequence.String());
   return ret;
}

filter_result KeyStrokeSequenceCommandActuator :: KeyEvent(const BMessage * keyMsg, BList * outlist, void ** setAsyncData, BMessage * lastMouseMove)
{
   if (IS_KEY_DOWN(keyMsg))
   {
      BMessage temp(*keyMsg);
      int numChars = _sequence.Length();
      for (int i=0; i<numChars; i++)
      {
         char nextChar = _sequence.ByteAt(i);
 
         temp.RemoveName("modifiers");
         temp.AddInt32("modifiers", _modCodes[i]);
      
         uint32 keyCode = _keyCodes[i];
         temp.RemoveName("key");  
         if (keyCode >= 0) temp.AddInt32("key", keyCode); 
      
         temp.RemoveName("raw_char");
         temp.AddInt32("raw_char", (int32) nextChar);
         
         temp.RemoveName("byte");
         temp.AddInt8("byte", nextChar);
         
         temp.RemoveName("states");
         temp.AddData("states", B_UINT8_TYPE, &_states[i*16], 16, true, 16);
      
         char t[2];
         t[0] = nextChar;
         t[1] = '\0';
         temp.ReplaceString("bytes", t);
      
         temp.what = B_KEY_DOWN;
         outlist->AddItem(new BMessage(temp));

         temp.what = B_KEY_UP;
         outlist->AddItem(new BMessage(temp));      
      }
      return B_DISPATCH_MESSAGE;
   }
   else return B_SKIP_MESSAGE;
}


BArchivable * KeyStrokeSequenceCommandActuator ::Instantiate(BMessage * from)
{
   return (validate_instantiation(from, "KeyStrokeSequenceCommandActuator")) ? new KeyStrokeSequenceCommandActuator(from) : NULL;
}



///////////////////////////////////////////////////////////////////////////////
//
//  MIMEHandlerCommandActuator
//
///////////////////////////////////////////////////////////////////////////////

MIMEHandlerCommandActuator :: MIMEHandlerCommandActuator(int32 argc, char ** argv)
   : CommandActuator(argc, argv), _mimeType((argc > 1) ? argv[1] : "")
{
   // empty
}

MIMEHandlerCommandActuator :: MIMEHandlerCommandActuator(BMessage * from)
   : CommandActuator(from)
{
   const char * temp;
   if (from->FindString("mimeType", 0, &temp) == B_NO_ERROR) _mimeType = temp;   
}

MIMEHandlerCommandActuator :: ~MIMEHandlerCommandActuator()
{
   // empty
}

status_t MIMEHandlerCommandActuator :: Archive(BMessage * into, bool deep) const
{
   status_t ret = CommandActuator :: Archive(into, deep);
   into->AddString("mimeType", _mimeType.String());
   return ret;
}

filter_result MIMEHandlerCommandActuator :: KeyEvent(const BMessage * keyMsg, BList * outlist, void ** setAsyncData, BMessage * lastMouseMove)
{
   if (IS_KEY_DOWN(keyMsg)) *setAsyncData = (void *) true;  // cause KeyEventAsync() to be called asynchronously
   return B_SKIP_MESSAGE;
}

void MIMEHandlerCommandActuator :: KeyEventAsync(const BMessage * keyMsg, void * asyncData)
{
   if (be_roster)
   {
      status_t  ret = be_roster->Launch(_mimeType.String());
      if ((ret != B_NO_ERROR)&&(ret != B_ALREADY_RUNNING))
      {
         const char * str1 = "Can't launch handler for ";
         const char * str2 = ", no such MIME type exists.  Please check your SpicyKeys settings.";
         char * temp = new char[strlen(str1) + strlen(_mimeType.String()) + strlen(str2) + 1];
         sprintf(temp, "%s%s%s", str1, _mimeType.String(), str2);
         (new BAlert("SpicyKeys MIME Launcher Error", temp, "Okay"))->Go(NULL); 
         delete [] temp;      
      }
   }
}


BArchivable * MIMEHandlerCommandActuator ::Instantiate(BMessage * from)
{
   return (validate_instantiation(from, "MIMEHandlerCommandActuator")) ? new MIMEHandlerCommandActuator(from) : NULL;
}



///////////////////////////////////////////////////////////////////////////////
//
//  BeepCommandActuator
//
///////////////////////////////////////////////////////////////////////////////

BeepCommandActuator :: BeepCommandActuator(int32 argc, char ** argv)
   : CommandActuator(argc, argv)
{
   // empty
}

BeepCommandActuator :: BeepCommandActuator(BMessage * from)
   : CommandActuator(from)
{
   // empty
}

BeepCommandActuator :: ~BeepCommandActuator()
{
   // empty
}

status_t BeepCommandActuator :: Archive(BMessage * into, bool deep) const
{
   status_t ret = CommandActuator :: Archive(into, deep);
   return ret;
}

BArchivable * BeepCommandActuator ::Instantiate(BMessage * from)
{
   return (validate_instantiation(from, "BeepCommandActuator")) ? new BeepCommandActuator(from) : NULL;
}

filter_result BeepCommandActuator :: KeyEvent(const BMessage * keyMsg, BList * outlist, void ** setAsyncData, BMessage * lastMouseMove)
{
   if (IS_KEY_DOWN(keyMsg)) beep();
   return B_SKIP_MESSAGE;
}

///////////////////////////////////////////////////////////////////////////////
//
//  MultiCommandActuator
//
///////////////////////////////////////////////////////////////////////////////

MultiCommandActuator :: MultiCommandActuator(BMessage * from)
   : CommandActuator(from)
{
   BMessage msg;
   for (int i=0; from->FindMessage("subs", i, &msg) == B_NO_ERROR; i++)
   {
      BArchivable * subObj = instantiate_object(&msg);
      if (subObj)
      {
         CommandActuator * ca = dynamic_cast<CommandActuator*>(subObj);
         if (ca) _subActuators.AddItem(ca);
            else delete subObj;
      }
   }
}

MultiCommandActuator :: MultiCommandActuator(int32 argc, char ** argv)
   : CommandActuator(argc, argv)
{
   for (int i=1; i<argc; i++)
   {
      CommandActuator * sub = CreateCommandActuator(argv[i]);
      if (sub) _subActuators.AddItem(sub);
          else printf("Error creating subActuator from [%s]\n", argv[i]);
   }
}

MultiCommandActuator :: ~MultiCommandActuator()
{
   int numSubs = _subActuators.CountItems();
   for (int i=0; i<numSubs; i++) delete ((CommandActuator *) _subActuators.ItemAt(i));
}

status_t MultiCommandActuator :: Archive(BMessage * into, bool deep) const
{
   status_t ret = CommandActuator :: Archive(into, deep);
   if (ret != B_NO_ERROR) return ret;
   
   int numSubs = _subActuators.CountItems();
   for (int i=0; i<numSubs; i++)
   {
      BMessage msg;
      ret = ((CommandActuator *) _subActuators.ItemAt(i))->Archive(&msg, deep);
      if (ret != B_NO_ERROR) return ret;
      into->AddMessage("subs", &msg);
   }
   return B_NO_ERROR;
}

BArchivable * MultiCommandActuator ::Instantiate(BMessage * from)
{
   return (validate_instantiation(from, "MultiCommandActuator")) ? new MultiCommandActuator(from) : NULL;
}

filter_result MultiCommandActuator :: KeyEvent(const BMessage * keyMsg, BList * outlist, void ** asyncData, BMessage * lastMouseMove)
{
   BList * aDataList = NULL;   // demand-allocated
   filter_result res = B_SKIP_MESSAGE;
   int numSubs = _subActuators.CountItems();
   for (int i=0; i<numSubs; i++)
   {
      void * aData = NULL;
      status_t next = ((CommandActuator *) _subActuators.ItemAt(i))->KeyEvent(keyMsg, outlist, &aData, lastMouseMove);
      if (next == B_DISPATCH_MESSAGE) res = B_DISPATCH_MESSAGE;  // dispatch message if at least one sub wants it dispatched
      if (aData)
      {
         if (aDataList == NULL) *asyncData = aDataList = new BList;
         while(aDataList->CountItems() < i-1) aDataList->AddItem(NULL);
         aDataList->AddItem(aData);
      }
   }

#if DEBUG_CRAP
BFile dbg("/boot/home/debug.msg", B_CREATE_FILE|B_WRITE_ONLY|B_OPEN_AT_END);
char buf[50]; sprintf(buf, "%i", outlist->CountItems()); (new BAlert(buf,buf,buf))->Go(NULL);
for (int i=0; i<outlist->CountItems(); i++)
{
   BMessage * m = (BMessage *) outlist->ItemAt(i);
   m->Flatten(&dbg);
}
#endif

   return res;
}

void MultiCommandActuator :: KeyEventAsync(const BMessage * keyUpMsg, void * asyncData)
{
   BList * list = (BList *) asyncData;
   int numSubs = list->CountItems();  // list may be shorter than _subActuators.CountItems(), but it won't be longer
   for (int i=0; i<numSubs; i++)
   {
      void * aData = list->ItemAt(i);
      if (aData) ((CommandActuator *) _subActuators.ItemAt(i))->KeyEventAsync(keyUpMsg, aData);
   }
   delete list;
}

///////////////////////////////////////////////////////////////////////////////
//
//  MoveMouseCommandActuator
//
///////////////////////////////////////////////////////////////////////////////

MoveMouseCommandActuator :: MoveMouseCommandActuator(BMessage * from)
   : CommandActuator(from)
{
   if (from->FindFloat("xPercent", &_xPercent) != B_NO_ERROR) _xPercent = 0.0f;
   if (from->FindFloat("yPercent", &_yPercent) != B_NO_ERROR) _yPercent = 0.0f;
   if (from->FindFloat("xPixels",  &_xPixels)  != B_NO_ERROR) _xPixels = 0;
   if (from->FindFloat("yPixels",  &_yPixels)  != B_NO_ERROR) _yPixels = 0;
}

MoveMouseCommandActuator :: MoveMouseCommandActuator(int32 argc, char ** argv)
   : CommandActuator(argc, argv), _xPercent(0.0f), _yPercent(0.0f), _xPixels(0), _yPixels(0)
{
   if (argc > 1) ParseArg(argv[1], _xPercent, _xPixels);
   if (argc > 2) ParseArg(argv[2], _yPercent, _yPixels);
}

MoveMouseCommandActuator :: ~MoveMouseCommandActuator()
{
   // empty
}

status_t MoveMouseCommandActuator :: Archive(BMessage * into, bool deep) const
{
   status_t ret = CommandActuator :: Archive(into, deep);
   into->AddFloat("xPercent", _xPercent);
   into->AddFloat("yPercent", _yPercent);
   into->AddFloat("xPixels", _xPixels);
   into->AddFloat("yPixels", _yPixels);
   return ret;
}

void MoveMouseCommandActuator :: CalculateCoords(float & setX, float & setY) const
{
   BScreen s;
   BRect frame = s.Frame();
   setX = (frame.Width()  * _xPercent) + _xPixels;
   setY = (frame.Height() * _yPercent) + _yPixels;
}

BMessage * MoveMouseCommandActuator :: CreateMouseMovedMessage(const BMessage * origMsg, BPoint p, BList * outlist) const
{
   // Force p into the screen space
   {
      BScreen s;
      p.ConstrainTo(s.Frame());
   }

   BMessage * newMsg = new BMessage(B_MOUSE_MOVED);

   newMsg->AddPoint("where", p);

   int32 buttons = 0;
   (void)origMsg->FindInt32("buttons", &buttons);
if (buttons == 0) buttons = 1;
   newMsg->AddInt32("buttons", buttons);
   
   // Trey sez you gotta keep then "when"'s increasing if you want click&drag to work!
   const BMessage * lastMessage = (outlist->CountItems() > 0) ? (const BMessage *)outlist->ItemAt(outlist->CountItems()-1) : origMsg;
   int64 when;
   if (lastMessage->FindInt64("when", &when) == B_NO_ERROR)
   {
      when++;
      newMsg->RemoveName("when");
      newMsg->AddInt64("when", when);
   }   
   
   return newMsg;
}

static bool IsNumeric(char c);
static bool IsNumeric(char c) {return (((c >= '0')&&(c <= '9'))||(c == '.')||(c == '-'));}

// Parse a string of the form "10", "10%", "10+10%", or "10%+10"
void MoveMouseCommandActuator :: ParseArg(const char * arg, float & setPercent, float & setPixels) const
{
   char * temp = new char[strlen(arg)+1];
   strcpy(temp, arg);
   
   // Find the percent part, if any
   char * percent = strchr(temp, '%');
   if (percent)
   {
      // Rewind to one before the beginning of the number
      char * beginNum = percent - 1;
      while(beginNum >= temp)
      {
         char c = *beginNum;
         if (IsNumeric(c)) beginNum--;
                      else break;
      }
      
      // parse the number
      setPercent = atof(++beginNum)/100.0f;

      // Now white it out to ease finding the other #
      while(beginNum <= percent) *(beginNum++) = ' ';
   }
   
   // Find the pixel part, if any
   char * pixel = temp;
   while(!IsNumeric(*pixel)) 
   {
      if (*pixel == '\0') break;
      pixel++;
   }
   setPixels = atof(pixel);
   
   delete [] temp;
}

///////////////////////////////////////////////////////////////////////////////
//
//  MoveMouseToCommandActuator
//
///////////////////////////////////////////////////////////////////////////////

MoveMouseToCommandActuator :: MoveMouseToCommandActuator(BMessage * from)
   : MoveMouseCommandActuator(from)
{
   // empty
}

MoveMouseToCommandActuator :: MoveMouseToCommandActuator(int32 argc, char ** argv)
   : MoveMouseCommandActuator(argc, argv)
{
   // empty
}

MoveMouseToCommandActuator :: ~MoveMouseToCommandActuator()
{
   // empty
}

status_t MoveMouseToCommandActuator :: Archive(BMessage * into, bool deep) const
{
   status_t ret = MoveMouseCommandActuator :: Archive(into, deep);
   return ret;
}


BArchivable * MoveMouseToCommandActuator ::Instantiate(BMessage * from)
{
   return (validate_instantiation(from, "MoveMouseToCommandActuator")) ? new MoveMouseToCommandActuator(from) : NULL;
}


filter_result MoveMouseToCommandActuator :: KeyEvent(const BMessage * keyMsg, BList * outlist, void ** setAsyncData, BMessage * lastMouseMove)
{
   if (IS_KEY_DOWN(keyMsg))
   {
      float x, y;
      CalculateCoords(x, y);      
      BMessage * newMsg = CreateMouseMovedMessage(keyMsg, BPoint(x, y), outlist);
      *lastMouseMove = *newMsg;
      outlist->AddItem(newMsg);
      return B_DISPATCH_MESSAGE;
   }
   return B_SKIP_MESSAGE;
}


///////////////////////////////////////////////////////////////////////////////
//
//  MoveMouseByCommandActuator
//
///////////////////////////////////////////////////////////////////////////////

MoveMouseByCommandActuator :: MoveMouseByCommandActuator(BMessage * from)
   : MoveMouseCommandActuator(from)
{
   // empty
}

MoveMouseByCommandActuator :: MoveMouseByCommandActuator(int32 argc, char ** argv)
   : MoveMouseCommandActuator(argc, argv)
{
   // empty
}

MoveMouseByCommandActuator :: ~MoveMouseByCommandActuator()
{
   // empty
}

status_t MoveMouseByCommandActuator :: Archive(BMessage * into, bool deep) const
{
   status_t ret = MoveMouseCommandActuator :: Archive(into, deep);
   return ret;
}


BArchivable * MoveMouseByCommandActuator ::Instantiate(BMessage * from)
{
   return (validate_instantiation(from, "MoveMouseByCommandActuator")) ? new MoveMouseByCommandActuator(from) : NULL;
}


filter_result MoveMouseByCommandActuator :: KeyEvent(const BMessage * keyMsg, BList * outlist, void ** setAsyncData, BMessage * lastMouseMove)
{
   if (IS_KEY_DOWN(keyMsg))
   {
      // Get the current mouse position
      BPoint p;
      if (lastMouseMove->FindPoint("where", &p) == B_NO_ERROR) 
      {
         // Get the desired offset
         BPoint diff;
         CalculateCoords(diff.x, diff.y);      
         p += diff;
   
         BMessage * newMsg = CreateMouseMovedMessage(keyMsg, p, outlist);
         *lastMouseMove = *newMsg;
         outlist->AddItem(newMsg);
         return B_DISPATCH_MESSAGE;
      }
   }
   return B_SKIP_MESSAGE;
}


///////////////////////////////////////////////////////////////////////////////
//
//  SendMessageCommandActuator
//
///////////////////////////////////////////////////////////////////////////////

SendMessageCommandActuator :: SendMessageCommandActuator(int32 argc, char ** argv)
   : CommandActuator(argc, argv), _signature((argc > 1) ? argv[1] : "")
{
   // parse the what code.  It may be in any of the following formats:
   //  12356    (int)
   //  'HELO'   (chars enclosed in single quotes)
   //  0x12ab3c (hex)

   if (argc > 2) 
   {
      const char * whatStr = argv[2];
      if ((whatStr[0] == '\'')&&(strlen(whatStr) == 6)&&(whatStr[5] == '\''))
      {
         // Translate the characters into the uint32 they stand for.
         // Note that we must do this in a byte-endian-independant fashion (no casting!)
         _sendMsg.what = 0;
         uint32 mult = 1;
         for (int i=0; i<4; i++)
         {
            _sendMsg.what += ((uint32)(whatStr[4-i])) * mult;
            mult <<= 8;
         }
      }
      else if (strncmp(whatStr, "0x", 2) == 0) _sendMsg.what = strtoul(&whatStr[2], NULL, 16);  // translate hex string to decimal
      else _sendMsg.what = atoi(whatStr);
   }
   else _sendMsg.what = 0;

   // parse any fields
   // format:
   //   name               (bool, true)
   //   name=value         (string)
   //   name:string=value  (string)
   //   name:int8=value    (int8)
   //   name:int16=value   (int16)
   //   name:int32=value   (int32)
   //   name:int64=value   (int64)
   //   name:bool=value    (bool)
   //   name:float=value   (float)
   //   name:double=value  (double)
   //   name:point=value   (point)
   //   name:rect=value    (rect)

   for (int i=3; i<argc; i++)
   {
      type_code tc = B_BOOL_TYPE;  // default type when no value is present
      const char * arg = argv[i];
      BString argStr(arg);
      const char * equals = strchr(arg, '=');
      const char * value = "true";  // default if no value is present
      if (equals)
      {
         tc = B_STRING_TYPE;  // default type when value is present
         value = equals+1;
         const char * colon = strchr(arg, ':');
         if (colon > equals) colon = NULL;  // colons after the equals sign don't count
         if (colon)
         {
            const char * typeStr = colon+1; 
                 if (strncasecmp(typeStr, "string", 6) == 0) tc = B_STRING_TYPE;
            else if (strncasecmp(typeStr, "int8",   4) == 0) tc = B_INT8_TYPE;
            else if (strncasecmp(typeStr, "int16",  5) == 0) tc = B_INT16_TYPE;
            else if (strncasecmp(typeStr, "int32",  5) == 0) tc = B_INT32_TYPE;
            else if (strncasecmp(typeStr, "int64",  5) == 0) tc = B_INT64_TYPE;
            else if (strncasecmp(typeStr, "bool",   4) == 0) tc = B_BOOL_TYPE;
            else if (strncasecmp(typeStr, "float",  5) == 0) tc = B_FLOAT_TYPE;
            else if (strncasecmp(typeStr, "double", 6) == 0) tc = B_DOUBLE_TYPE;
            else if (strncasecmp(typeStr, "point",  5) == 0) tc = B_POINT_TYPE;
            else if (strncasecmp(typeStr, "rect",   4) == 0) tc = B_RECT_TYPE;
            argStr = argStr.Truncate(colon-arg);  // remove the colon and stuff
         }
         else argStr = argStr.Truncate(equals-arg);  // remove the equals and arg
      }

      switch(tc)
      {
         case B_STRING_TYPE: _sendMsg.AddString(argStr.String(), value);                              break;
         case B_INT8_TYPE:   _sendMsg.AddInt8(argStr.String(),   (int8)atoi(value));                  break;
         case B_INT16_TYPE:  _sendMsg.AddInt16(argStr.String(),  (int16)atoi(value));                 break;
         case B_INT32_TYPE:  _sendMsg.AddInt32(argStr.String(),  (int32)atoi(value));                 break;
         case B_INT64_TYPE:  _sendMsg.AddInt64(argStr.String(),  (int64)atoi(value));                 break;
         case B_BOOL_TYPE:   _sendMsg.AddBool(argStr.String(),   ((value[0]=='t')||(value[0]=='T'))); break;
         case B_FLOAT_TYPE:  _sendMsg.AddFloat(argStr.String(),  atof(value));                        break;
         case B_DOUBLE_TYPE: _sendMsg.AddDouble(argStr.String(), (double)atof(value));                break;
         case B_POINT_TYPE: 
         {
            float pts[2] = {0.0f, 0.0f};
            ParseFloatArgs(pts, 2, value);
            _sendMsg.AddPoint(argStr.String(), BPoint(pts[0], pts[1])); 
         }
         break;
         case B_RECT_TYPE: 
         {
            float pts[4] = {0.0f, 0.0f, 0.0f, 0.0f};
            ParseFloatArgs(pts, 4, value);
            _sendMsg.AddRect(argStr.String(), BRect(pts[0], pts[1], pts[2], pts[3])); 
            break;
         }
      }
   }
//printf("sig=[%s]\n", _signature.String());
//_sendMsg.PrintToStream();
}


void
SendMessageCommandActuator :: 
ParseFloatArgs(float * args, int maxArgs, const char * str) const
{
   const char * next = str;
   for (int i=0; i<maxArgs; i++)
   {
      args[i] = atof(next);
      next = strchr(next, ',');
      if (next) next++;
           else break;
   }
}

SendMessageCommandActuator :: SendMessageCommandActuator(BMessage * from) : CommandActuator(from)
{
   const char * temp;
   if (from->FindString("signature", 0, &temp) == B_NO_ERROR) _signature = temp;
   (void) from->FindMessage("sendmsg", &_sendMsg);
}

SendMessageCommandActuator :: ~SendMessageCommandActuator()
{
   // empty
}

status_t SendMessageCommandActuator :: Archive(BMessage * into, bool deep) const
{
   status_t ret = CommandActuator :: Archive(into, deep);
   into->AddString("signature", _signature.String());
   into->AddMessage("sendmsg", &_sendMsg);
   return ret;
}

filter_result SendMessageCommandActuator :: KeyEvent(const BMessage * keyMsg, BList * outlist, void ** setAsyncData, BMessage * lastMouseMove)
{
   if (IS_KEY_DOWN(keyMsg)) *setAsyncData = (void *) true;  // cause KeyEventAsync() to be called asynchronously
   return B_SKIP_MESSAGE;
}

void SendMessageCommandActuator :: KeyEventAsync(const BMessage * keyMsg, void * asyncData)
{
   if (be_roster)
   {
      if (_signature.Length() == 0) (new BAlert("SpicyKeys SendMessage Error", "SendMessage:  Target App Signature not specified", "Okay"))->Go(NULL);
      else
      {
         status_t error = B_OK;
         BMessenger msngr(_signature.String(), -1, &error);
         if (error == B_OK) msngr.SendMessage(&_sendMsg);
      }
   }
}

BArchivable * SendMessageCommandActuator ::Instantiate(BMessage * from)
{
   return (validate_instantiation(from, "SendMessageCommandActuator")) ? new SendMessageCommandActuator(from) : NULL;
}
