#include <stdio.h>
#include <kernel/OS.h>
#include <storage/File.h>
#include <storage/NodeMonitor.h>
#include <storage/Entry.h>
#include <game/WindowScreen.h>
#include <app/MessageFilter.h>
#include <support/Beep.h>

#include "SpicyKeysFilterConstants.h"
#include "KeyCommandMap.h"
#include "BitFieldTesters.h"
#include "CommandActuators.h"

#define J_FILE_UPDATED 'JfUp'

class hks 
{
public:
   hks(int32 key, BitFieldTester * t, CommandActuator * act, const BMessage & a)
      : _key(key), _tester(t), _actuator(act), _actuatorMessage(a)
   {
      // empty
   }
   
   ~hks() 
   {
      delete _actuator;
      delete _tester;
   }
   
   int32 GetKey() const {return _key;}
   bool DoModifiersMatch(uint32 bits) const {return _tester->isMatching(bits);}
   const BMessage & GetActuatorMsg() const {return _actuatorMessage;}
   CommandActuator * GetActuator() {return _actuator;}
   
private:
   int32 _key;
   BitFieldTester * _tester;
   CommandActuator * _actuator;
   const BMessage _actuatorMessage;
};

KeyCommandMap :: KeyCommandMap(const char * file) : _specs(NULL)
{
   _fileName = new char[strlen(file)+1];
   strcpy(_fileName, file);

   BEntry fileEntry(_fileName);
   if (fileEntry.InitCheck() == B_NO_ERROR)
   {
      node_ref nref;
      if (fileEntry.GetNodeRef(&nref) == B_NO_ERROR) watch_node(&nref, B_WATCH_STAT, this);
   }

   BMessage msg(J_FILE_UPDATED);
   PostMessage(&msg);

   _port = create_port(1, SPICYKEYS_CATCHER_PORT_NAME);
   PutMessageToPort();  // advertise our BMessenger to the world
}

KeyCommandMap :: ~KeyCommandMap()
{
   if (_port >= 0) close_port(_port);
   for (int i=_injects.CountItems()-1; i>=0; i--) delete (BMessage *)_injects.ItemAt(i);
   stop_watching(this);  // don't know if this is necessary, but it can't hurt
   DeleteHKSList(_specs);
   delete [] _fileName;
}

void KeyCommandMap :: MouseMoved(const BMessage * mm)
{
   // Save the mouse state for later...
   _lastMouseMessage = *mm;
}

filter_result KeyCommandMap :: KeyEvent(const BMessage * keyMsg, BList * outlist, const BMessenger & sendTo) 
{
   uint32 modifiers;
   filter_result ret = B_DISPATCH_MESSAGE;  // default:  pass it on
   
   if (keyMsg->FindInt32("modifiers", (int32 *) &modifiers) == B_NO_ERROR)
   {
      int32 key;
      if (keyMsg->FindInt32("key", &key) == B_NO_ERROR)
      {
         if (_syncSpecs.Lock())
         {
            if (_specs != NULL)
            {
               int num = _specs->CountItems();
               for (int i=0; i<num; i++)
               {
                  hks * next = (hks *) _specs->ItemAt(i);                  
   
                  if ((key == next->GetKey())&&(next->DoModifiersMatch(modifiers)))
                  {
                     void * asyncData = NULL;
                     ret = next->GetActuator()->KeyEvent(keyMsg, outlist, &asyncData, &_lastMouseMessage);
                     if (asyncData)
                     {
                        BMessage newMsg(*keyMsg);
                        newMsg.AddMessage("act", &next->GetActuatorMsg());
                        newMsg.AddPointer("adata", asyncData);
                        sendTo.SendMessage(&newMsg);
                     }
                  }
               }
            }  
            _syncSpecs.Unlock();
         }
      }
   }
   return ret;
}

void KeyCommandMap :: DrainInjectedEvents(const BMessage * keyMsg, BList * outlist, const BMessenger & sendTo)
{
   BList temp;
   if (_syncSpecs.Lock())
   {
      temp = _injects;
      _injects.MakeEmpty();
      _syncSpecs.Unlock();
   }
   int is = temp.CountItems();
   for (int i=0; i<is; i++) 
   {
      BMessage * msg = (BMessage *)temp.ItemAt(i);
      BArchivable * arc = instantiate_object(msg);
      if (arc)
      {
         CommandActuator * act = dynamic_cast<CommandActuator *>(arc);
         if (act)
         {
            BMessage newMsg(*keyMsg);
            newMsg.what = B_KEY_DOWN;  // it's what everyone expects, ya know
            void * asyncData = NULL;
            (void) act->KeyEvent(&newMsg, outlist, &asyncData, &_lastMouseMessage);
            if (asyncData)
            {
               newMsg.AddMessage("act", msg);
               newMsg.AddPointer("adata", asyncData);
               sendTo.SendMessage(&newMsg);
            }
         }
         delete arc;
      }
      delete msg;
   }
}

void KeyCommandMap :: MessageReceived(BMessage * msg)
{
   switch(msg->what)
   {
      case J_EXECUTE_COMMAND:
      {
         BMessage subMsg;
         if (msg->FindMessage("act", &subMsg) == B_NO_ERROR)
         {
            if (_syncSpecs.Lock())
            {
               _injects.AddItem(new BMessage(subMsg));
               _syncSpecs.Unlock();

               // This evil hack forces input_server to call Filter() on us so we can process the injected event.
               BPoint lmp;
               if (_lastMouseMessage.FindPoint("where", &lmp) == B_NO_ERROR) set_mouse_position((int)lmp.x, (int)lmp.y);
            }
         }
      }
      break;

      case J_REPLENISH_MESSENGER:
         PutMessageToPort();
      break;

      case B_NODE_MONITOR:
         /* fall through to J_FILE_UPDATED */         
      case J_FILE_UPDATED:
      {  
         BMessage fileMsg;
         BFile file(_fileName, B_READ_ONLY);         
         if ((file.InitCheck() == B_NO_ERROR)&&(fileMsg.Unflatten(&file) == B_NO_ERROR))
         {
            BList * newList = new BList;
            BList * oldList = NULL;  // whatever this is set to will be deleted below -- defaults to no deletion
            
            int i=0;
            BMessage nextMsg;
            while(fileMsg.FindMessage("spec", i++, &nextMsg) == B_NO_ERROR) 
            {
               uint32 key;
               BMessage testerMsg;
               BMessage actMsg;
               
               if ((nextMsg.FindInt32("key", (int32 *) &key)       == B_NO_ERROR) &&
                   ((nextMsg.FindMessage("act", &actMsg))          == B_NO_ERROR) &&
                   ((nextMsg.FindMessage("modtester", &testerMsg)) == B_NO_ERROR))
               {
                  BArchivable * arcObj = instantiate_object(&testerMsg);
                  if (arcObj)
                  {
                     BitFieldTester * tester = dynamic_cast<BitFieldTester*>(arcObj);
                     if (tester)
                     {
                         BArchivable * barcObj = instantiate_object(&actMsg);
                         if (barcObj)
                         {
                            CommandActuator * act = dynamic_cast<CommandActuator*>(barcObj);
                            if (act) newList->AddItem(new hks(key, tester, act, actMsg));
                            else 
                            {
                               delete barcObj;
                               delete tester;
                            }
                         }
                         else delete tester;
                     }
                     else delete arcObj;
                  }
               }
            }
            
            if (_syncSpecs.Lock()) 
            {
               // swap in the new list
               oldList = _specs;
               _specs = newList;
               _syncSpecs.Unlock();
            }
            else 
            {
               // wtf?  This shouldn't happen...
               oldList = newList;  // but clean up if it does
            }            
            DeleteHKSList(oldList);
         }
      }
      break;
   }
}

// Deletes an HKS-filled BList and its contents.
void KeyCommandMap :: DeleteHKSList(BList * l)
{
   if (l != NULL)
   {
      int num = l->CountItems();
      for (int i=0; i<num; i++) delete ((hks *) l->ItemAt(i));
      delete l;
   }
}

void KeyCommandMap :: PutMessageToPort()
{
   if (_port >= 0)
   {
      BMessenger toMe(this);
      BMessage m;
      m.AddMessenger("target", toMe);

      char buf[2048];
      ssize_t fs = m.FlattenedSize();
      if ((fs <= sizeof(buf))&&(m.Flatten(buf, fs) == B_NO_ERROR)) write_port_etc(_port, 0, buf, fs, B_TIMEOUT, 250000);
   }
}
